/*
								+---------------------------------+
								|                                 |
								|  ***   Tiny jpeg encoder   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            JpegEnc.h            |
								|                                 |
								+---------------------------------+
*/

#ifndef __BMP_INCLUDED
#define __BMP_INCLUDED

struct TBmp {
	char n_former_bpp; // former bpp, before conversion to RGBA8
	bool b_grayscale;
	bool b_alpha;
	int n_width, n_height;
	unsigned __int32 *p_buffer;
};

#endif //__BMP_INCLUDED

#ifndef __JPEG_ENC_INCLUDED
#define __JPEG_ENC_INCLUDED

/*
 *		=== config section ===
 */

//#define __JPEG_ENC_VERBOSE
// print some info on what is being written (to stdio; involves various kinds of tables, etc)

//#define __JPEG_ENC_ENCODE_VERBOSE
// print encoding progress to stdio (source image window, DCT + quantized, RLE compressed)

//#define __JPEG_ENC_HUFF_ENCODE_VERBOSE
// print huffman encoding stage progress to stdio (RLE data + it's huffman representation)

//#define __JPEG_ENC_FAST_CHROMINANCE_SAMPLING
// if defined, chrominance will be sampled just from a single pixel
// it can save some time when encoding images with 2x2 or 4x4 sampling,
// but the quality can be lower, that depends on type of the image, of course
// resulting code is smaller if enabled

//#define __JPEG_ENC_LOOKUP_CHUFFCODER
// use lookup table for symbol encoding instead of CHuffmanCoder class (a bit (cca 9%) faster
// but uses some more memory on stack and propably more memory on heap, code is a few bytes larger)

//#define __JPEG_ENC_USE_REFERENCE_DCT
// DCT debugging (do not use with conjunction with __JPEG_ENC_INTEGER_FDCT, it won't compile)

#ifndef __JPEG_ENC_USE_REFERENCE_DCT
//#define __JPEG_ENC_INTEGER_FDCT
#endif
// use faster integer DCT (on my machine, it's actualy not faster, but it should be - i guess)

#define __JPEG_ENC_ENABLE_LOSSLESS
// if defined, lossless compression algorithm is also available

#ifdef __JPEG_ENC_ENABLE_LOSSLESS
#define __JPEG_ENC_USE_LOSSLESS_NONSTD_RLE
#endif
// if defined, jpeg lossless algorithm is further enhanced by RLE compression this is non-standard
// extension which would propably make std decoder crash or at least to decode rubbish
// never use in programs where your jpeg files are available in standalone-file form so they could
// be potentialy copied away! do not breed nor allow anyone to breed non-standard jpegs away from
// your app (note there is no record in image file header wheter this extension was used or not)

//#define __JPEG_ENC_INTEGER_QUANT
// integer quant table coeffs (this actually is faster; lower image quality, introduces quite
// a lot of noise; can be used without __JPEG_ENC_INTEGER_FDCT)

//#define __JPEG_ENC_WRITE_SINGLE_QUANT_TABLE_PER_DQT_BLOCK
// if defined, every quantization table is written into separate data-block
// (specs allow both separate and single-block versions, it's here only for compatibility reasons)

//#define __JPEG_ENC_WRITE_SINGLE_HUFF_TABLE_PER_DQT_BLOCK
// if defined, every huffman table is written into separate data-block
// (specs allow both separate and single-block versions, it's here only for compatibility reasons)

//#define __JPEG_ENC_ENABLE_CMYK
// if defined, it's possible to save jpegs, containing CMYK values instead of Y[CbCr]

//#define __JPEG_ENC_ENABLE_YCCK
// if defined, it's possible to save jpegs, containing YCCK values instead of Y[CbCr]

#define __JPEG_ENC_BGR
// if defined, swap color channels while encoding colors

/*
 *		=== ~config section ===
 */

class CQuantTableFactory {
public:
	/*
	 *	void CQuantTableFactory::Calc_QuantTable(unsigned char (&r_p_output)[64],
	 *		bool b_luminance, float f_quality)
	 *		- calculate quantization table, based on simple formula
	 *		- r_p_output is reference to array of 64 unsigned chars where table
	 *		  coefficients are to be output (in natural left-to-right top-to-bottom order, no zig-zag)
	 *		- set b_luminance to true in case the table should be used for quantizing luminance
	 *		  coefficients (table is a bit softer in that case as human eye is more sensitive
	 *		  to luminance errors)
	 *		- f_quality is in range 0 - 1.1 (can be negative up to approx -1.97647 which is limit)
	 */
	static void Calc_QuantTable(unsigned char (&r_p_output)[64], bool b_luminance, float f_quality);
};

class CJpegEncoder {
protected:
	static const char m_p_zig_indices[63];

	class COutputFile {
	protected:
		FILE *m_p_fw;

	public:
		COutputFile(FILE *p_fw);
		bool Write_Byte(unsigned char n_value);
		bool Write_Short(short n_value);
	};

	class CBitWriter {
	protected:
		unsigned char m_n_byte;
		int m_n_bit_num;

	public:
		CBitWriter();
		bool WriteBit(bool b_bit, COutputFile *p_output_file);
		bool WriteNumber(int n_bit_num, int n_value, COutputFile *p_output_file);
		bool PaddByte(COutputFile *p_output_file);
	};

	struct TQuantTable { // quantization table
		bool b_write; // wheter to write quant table to the file
#ifdef __JPEG_ENC_INTEGER_QUANT
		int p_value[64]; // 16:16 fixed-point
#else
		float p_value[64]; // pre-scaled values for AA&N FDCT
#endif
		unsigned char p_store_value[64]; // original values to be written to file
	} m_p_quant_table[2];

	struct THuffmanTable {
		unsigned char p_code_num[16]; // number of codes for all possible 16 lengths
		unsigned char *p_code[16]; // 16 pointers into p_code_table where the actual data are
		unsigned char p_code_table[256]; // table for all huffman codes
	};

	struct TRLEBlockData {
		struct TRLEPair {
			unsigned char n_code_word;
			// upper nibble is zero num, lower nibble is n_unsigned_coeff bit num
			short n_coeff;

			inline TRLEPair() {}
			inline TRLEPair(short n_x); // set n_code_word to code length (preceding zero num has yet to be shifted << 4 and or-ed)
			static inline unsigned char n_SignificantBit_Num(short n_x); // return number of significant bits in n_x (n_x as-is, no sign-dependent ops involved)
		};
		TRLEPair p_pair[64];
		unsigned char n_pair_num;
	};
	// data of a single compressed block, output by DCT and ready for huffman coding
	// (need to gather all codewords to generate optimized huffman table)

	struct TMacroBlock {
		TRLEBlockData *p_block;
	};
	// macro-block, can contain multiple blocks, pointers point to a single long
	// array -> do not free, free the array

	struct TComponent {
		unsigned char n_component_id;
		unsigned char n_sampling_horiz, n_sampling_vert;
		unsigned char n_scale_horiz, n_scale_vert;
		unsigned char n_quant_table_index;
		unsigned char n_dc_huff_coder_index, n_ac_huff_coder_index;
	};
	// component info (generated from encode color type)

	typedef CHuffmanCoder<16, unsigned char> CJpegHuffCoder; // sick of typing it again and again

	struct TBlockInfo {
		const TComponent *p_component;
		unsigned char n_offset_x, n_offset_y;
		TQuantTable *p_quant_table;
		CJpegHuffCoder *p_ac_huff_coder, *p_dc_huff_coder;
		int *p_ac_huff_code_table, *p_dc_huff_code_table;
	};
	// block info (used to speed-up compression, array with TBlockInfo for every block in
	// macro-block is generated, this array is then looped trough during encoding process)

public:
	enum {
		color_Gray,		// grayscale (Y = 0.299R + 0.587G + 0.114B, single huffman and quant table)
		color_RG,		// red-green (no color transform, single huffman and quant table, red is sampled as required, green is always sampled 1x1 (i.e. there is potentialy less green samples))
		color_RGB		// RGB (stored as YCbCr, two huffman and quant tables, one for Y, second for CbCr)
#ifdef __JPEG_ENC_ENABLE_CMYK
		, color_CMYK	// RGB (stored as CMYK, two tables, one for C and K, second for M and Y; note 4-component JPEG uses YCCK)
#endif
#ifdef __JPEG_ENC_ENABLE_YCCK
		, color_YCCK	// RGB (stored as YCbCrK, two tables, one for Y and K, second for Cb and Cr)
#endif
	};
	// encode color type

	/*
	 *	CJpegEncoder()
	 *		- default constructor
	 */
	CJpegEncoder();

	/*
	 *	void Write_QuantTable(int n_index, bool b_write)
	 *		- set write flag of quantization table with index n_index (n_index is 0 or 1)
	 *		- by default, both quantization tables are written (default for RGB images)
	 */
	void Write_QuantTable(int n_index, bool b_write);

	/*
	 *	void Set_QuantTable(int n_index, int n_color_component, const unsigned char p_quant_table[64])
	 *		- set data of quantization table with index n_index (n_index is 0 or 1)
	 *		- p_quant_table is pointer to 64 element array, containing quantization coeffs
	 *		  (coming in natural left-to-right, top-to-bottom order, no zig-zag)
	 */
	void Set_QuantTable(int n_index, const unsigned char p_quant_table[64]);

	/*
	 *	bool CJpegEncoder::Encode_Jpeg(FILE *p_fw, const TBmp *p_bitmap, int n_encode_color,
	 *		int n_color_sampling_horiz, int n_color_sampling_vert)
	 *		- encode jpeg image from p_bitmap (contains RGB(A) 8bpp image) to p_fw
	 *		  (p_fw has to be opened for binary writing; fwrite() only is used)
	 *		- bitmap resolution in jpeg is written as unsigned short so it's limited to 65535 ^ 2
	 *		- n_encode_color (one of color_Gray, color_RG, color_RGB, color_CMYK or color_YCCK)
	 *		  determines way of storing color in jpeg file (source color is RGB (A is never used))
	 *		- n_color_sampling_horiz and n_color_sampling_vert is color sampling (has to be 1, 2 or 4)
	 *		- for color_Gray, color sampling is forced to 1 x 1 (recommended in jpeg specification)
	 *		  and quant table 0 is used only
	 *		- for color_RG, R has dense 1 x 1 samples, G has samples of size n_color_sampling_horiz x
	 *		  n_color_sampling_vert, quant table 0 is used only
	 *		- color_RGB sets storing color as YCbCr, Y has dense samples and uses quant table 0,
	 *		  Cb and Cr has normal samples and both uses quant table 1
	 *		- color_CMYK (valid only if __JPEG_ENC_ENABLE_CMYK is defined) sets storing color
	 *		  information as CMYK, C and K has dense samples and uses quant table 0, M and Y has
	 *		  normal samples and uses quant table 1
	 *		- color_YCCK is used for standard 4-component jpegs (YCCK means YCbCrK) C and K has dense
	 *		  samples and uses quant table 0, M and Y has normal samples and uses quant table 1
	 *		- quant tables are by default set to be written, but their contents are unspecified
	 *		- returns true on success and false on failure (not enough memory for temporal info
	 *		  or not enough space on dest drive)
	 */
	bool Encode_Jpeg(FILE *p_fw, const TBmp *p_bitmap, int n_encode_color = color_RGB,
		int n_color_sampling_horiz = 2, int n_color_sampling_vert = 2);

#ifdef __JPEG_ENC_ENABLE_LOSSLESS
	/*
	 *	bool CJpegEncoder::Encode_LosslessJpeg(FILE *p_fw, const TBmp *p_bitmap,
	 *		int n_sample_precission = 8, bool b_separate_entropy = false, int n_channel_num = 3,
	 *		 int n_channel0_index = 0, int n_channel1_index = 1, int n_channel2_index = 2,
	 *		 int n_channel3_index = 3)
	 *		- encode lossless jpeg image from p_bitmap (contains RGB(A) 8bpp image) to p_fw
	 *		  (p_fw has to be opened for binary writing; fwrite() only is used)
	 *		- bitmap resolution in jpeg is written as unsigned short so it's limited to 65535 ^ 2
	 *		- if b_separate_entropy is true, every component has it's own entropy table. otherwise
	 *		  color components (if present) share one table and alpha channel (if present)
	 *		  is compressed using another one
	 *		- n_sample_precission is destination data precission and can be 2 - 12 (but source
	 *		  data are 8 bpp so reasonable range is 2 - 8 bits per sample)
	 *		- n_channel_num is number of channels (can be 1 to 4)
	 *		- n_channel0_index trough n_channel4_index are indices of channels in source RGBA quad
	 *		  (formula to get channel from RGBA quad is: (n_rgba >> (8 * n_channel_index)) & 0xff
	 *		  therefore channel index can be 0 (red), 1 (green), 2 (blue) or 3 (alpha)
	 *		  in case __JPEG_ENC_BGR is defined, index with value 0 is set to 2 and vice-versa)
	 *		  if n_channel_num is 1, n_channel0_index is used only, the rest is ignored
	 *		  if n_channel_num is 2, n_channel0_index and n_channel1_index are used, etc ...
	 */
	bool Encode_LosslessJpeg(FILE *p_fw, const TBmp *p_bitmap,
		int n_sample_precission = 8, bool b_separate_entropy = false, int n_channel_num = 3,
		int n_channel0_index = 0, int n_channel1_index = 1, int n_channel2_index = 2,
		int n_channel3_index = 3);
#endif

protected:
	static void Get_ComponentInfo(int n_encode_color, int n_color_sampling_horiz,
		int n_color_sampling_vert, TComponent *p_component_info,
		int &r_n_component_num, int &r_n_huff_coder_num);
	void Get_BlockInfo(const TComponent *p_component, int n_component_num,
		CJpegHuffCoder *p_huff_coder_base, int *p_huff_code_table_base,
		TBlockInfo *p_block_info, int &r_n_block_num);

	template <class TPixelType, class CColorConvertor>
	static void Get_ImageBlock(TPixelType *p_dest_block, const unsigned __int32 *p_image,
		int n_scanline_width, int n_residual_width, int n_residual_height, int n_sample_width,
		int n_sample_height, const CColorConvertor &r_conv);

	template <class TPixelType>
	static void FDCT_Quantize_Block(short *p_quantized_block,
		const TPixelType *p_image_block, const TQuantTable *p_quant_table);

	static void RLE_Encode_Block(TRLEBlockData &r_block, const short *p_quantized_block);

	static bool Write_JFIFHeader(COutputFile *p_output_file);
	bool Write_QuantTables(COutputFile *p_output_file);
	static bool Write_HuffmanTables(COutputFile *p_output_file,
		const THuffmanTable *p_huffman_table, int n_table_num, bool b_write_dc_only = false);
	static bool Write_FrameHeader(COutputFile *p_output_file, unsigned short n_marker,
		int n_width, int n_height, int n_sample_precission, int n_component_num,
		const TComponent *p_component);
	static bool Write_StartScan(COutputFile *p_output_file, int n_component_num,
		const TComponent *p_component, unsigned char n_start_of_spectrum_predictor = 0,
		unsigned char n_end_of_spectrum = 63);
#ifndef __JPEG_ENC_LOOKUP_CHUFFCODER
	template <class CJpegHuffCoderClass>
	static THuffmanTable t_HuffmanTable(CJpegHuffCoderClass &r_huff_coder);
	static inline bool HuffCompress_Write_BlockData(const TRLEBlockData &r_t_block,
		CJpegHuffCoder *p_cur_dc_huff_coder, CJpegHuffCoder *p_cur_ac_huff_coder,
		CBitWriter &r_bit_writer, COutputFile *p_out_file);
	// sorted-tree symbols variant (sometimes doesn't work, don't know why yet)
#else
	template <class CJpegHuffCoderClass>
	static THuffmanTable t_HuffmanTable(CJpegHuffCoderClass &r_huff_coder, int *p_huff_code_table);
	static inline bool HuffCompress_Write_BlockData(const TRLEBlockData &r_t_block,
		const int *p_cur_dc_huff_code_table, const int *p_cur_ac_huff_code_table,
		CBitWriter &r_bit_writer, COutputFile *p_out_file);
#endif
};

#endif //__JPEG_ENC_INCLUDED
