/*
								+----------------------------------+
								|                                  |
								| ***   Texture loader class  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|          TextureUtil.h           |
								|                                  |
								+----------------------------------+
*/

/*
 *	2007-07-14
 *
 *	removed GL_TEXTURE_2D target from CGLTexture_2D constructor calls
 *
 */

#ifndef __TEXTURE_UTIL_INCLUDED
#define __TEXTURE_UTIL_INCLUDED

#include "Tga.h"
#include "Jpeg.h"

/*
 *	class CTextureLoader
 *		- simple class with a few useful functions for loading textures directly from file
 */
class CTextureLoader {
public:
	static CGLTexture_2D *p_LoadTexture_TGA(CGLState *p_state, const char *p_s_filename,
		bool b_mipmaps = true, bool b_clamp_to_edge = false)
	{
		TBmp *p_bitmap;

		if(!(p_bitmap = p_Load_TGA(p_s_filename)))
			return 0;

		CGLTexture_2D *p_texture;
		if(!(p_texture = new CGLTexture_2D(p_state, p_bitmap->n_width, p_bitmap->n_height,
		   GL_RGBA8, b_mipmaps, 0, GL_RGBA, GL_UNSIGNED_BYTE, p_bitmap->p_buffer))) {
			delete[] p_bitmap->p_buffer;
			delete p_bitmap;
			return 0;
		}
		delete[] p_bitmap->p_buffer;
		delete p_bitmap;

		if(b_clamp_to_edge) {
			p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
			p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
		}

		if(!p_texture->b_Status() || glGetError() != GL_NO_ERROR) {
			delete p_texture;
			return 0;
		}

		return p_texture;
	}

	static CGLTexture_2D *p_LoadTexture_Jpeg(CGLState *p_state, const char *p_s_filename,
		bool b_mipmaps = true, bool b_clamp_to_edge = false)
	{
		TBmp *p_bitmap;

		FILE *p_fr;
		if(!(p_fr = fopen(p_s_filename, "rb")))
			return 0;
		CJpegDecoder decoder;
		if(!(p_bitmap = decoder.p_Decode_JPEG(p_fr))) {
			fclose(p_fr);
			return 0;
		}
		fclose(p_fr);

		CGLTexture_2D *p_texture;
		if(!(p_texture = new CGLTexture_2D(p_state, p_bitmap->n_width, p_bitmap->n_height,
		   GL_RGBA8, b_mipmaps, 0, GL_BGRA, GL_UNSIGNED_BYTE, p_bitmap->p_buffer))) {
			delete[] p_bitmap->p_buffer;
			delete p_bitmap;
			return 0;
		}
		delete[] p_bitmap->p_buffer;
		delete p_bitmap;

		if(b_clamp_to_edge) {
			p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
			p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
		}

		if(!p_texture->b_Status() || glGetError() != GL_NO_ERROR) {
			delete p_texture;
			return 0;
		}

		return p_texture;
	}

	static CGLTexture_2D *p_LoadTexture_Hybrid(CGLState *p_state,
		const char *p_s_color_filename, const char *p_s_alpha_filename,
		bool b_mipmaps = true, bool b_clamp_to_edge = false)
	{
		TBmp *p_color_bitmap;
		FILE *p_fr;
		if(!(p_fr = fopen(p_s_color_filename, "rb")))
			return 0;
		CJpegDecoder decoder;
		if(!(p_color_bitmap = decoder.p_Decode_JPEG(p_fr))) {
			fclose(p_fr);
			return 0;
		}
		fclose(p_fr);
		// load RGB part

		TBmp *p_alpha_bitmap;
		if(!(p_alpha_bitmap = p_Load_TGA(p_s_alpha_filename))) {
			delete[] p_color_bitmap->p_buffer;
			delete p_color_bitmap;
			return 0;
		}
		// targa is supposed to contain alpha part

		int n_shift = (p_alpha_bitmap->b_alpha)? 0 : 8;
		// take alpha channel if present or red channel

		if(p_alpha_bitmap->n_width != p_color_bitmap->n_width ||
		   p_alpha_bitmap->n_height != p_color_bitmap->n_height) {
			delete[] p_color_bitmap->p_buffer;
			delete p_color_bitmap;
			delete[] p_alpha_bitmap->p_buffer;
			delete p_alpha_bitmap;
			return 0;
		}
		for(unsigned __int32 *p_alpha = p_alpha_bitmap->p_buffer,
		   *p_dest = p_color_bitmap->p_buffer, *p_end = p_color_bitmap->p_buffer +
		   (p_color_bitmap->n_width * p_color_bitmap->n_height); p_dest != p_end;
		   ++ p_dest, ++ p_alpha)
			*p_dest = (*p_dest & 0xffffff) | ((*p_alpha << n_shift) & 0xff000000);
		// merge bitmaps

		delete[] p_alpha_bitmap->p_buffer;
		delete p_alpha_bitmap;

		CGLTexture_2D *p_texture;
		if(!(p_texture = new CGLTexture_2D(p_state, p_color_bitmap->n_width,
		   p_color_bitmap->n_height, GL_RGBA8, b_mipmaps, 0, GL_BGRA,
		   GL_UNSIGNED_BYTE, p_color_bitmap->p_buffer))) {
			delete[] p_color_bitmap->p_buffer;
			delete p_color_bitmap;
			return 0;
		}
		delete[] p_color_bitmap->p_buffer;
		delete p_color_bitmap;

		if(b_clamp_to_edge) {
			p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
			p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
		}

		if(!p_texture->b_Status() || glGetError() != GL_NO_ERROR) {
			delete p_texture;
			return 0;
		}

		return p_texture;
	}
};

#endif //__TEXTURE_UTIL_INCLUDED
